/*******************************************************************************
 *
 *
 *    WashingtonDC Dreamcast Emulator
 *    Copyright (C) 2020, 2022, 2023, 2024 snickerbockers
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 ******************************************************************************/

#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>

#include "isofs.h"
#include "cdi.h"

#define CDI_V4 0x80000004
#define CDI_V5 0x80000005
#define CDI_V6 0x80000006

static void read_session(FILE *stream,
                         struct cdi_session *sess,
                         size_t *total_pos, unsigned ver);
static void read_track(FILE *stream,
                       struct cdi_track *track,
                       size_t *total_pos, unsigned ver);
static unsigned cdi_lba_base(void *ctxp);
static unsigned cdi_sector_size(void *ctxp);
static int cdi_read_block(void *ctxp, int block, void *buf, size_t buflen);

struct disc_intf const cdi_intf = {
    .base_lba = cdi_lba_base,
    .sector_size = cdi_sector_size,
    .read_block = cdi_read_block
};

cdi_file* cdi_load(char const *path) {
    cdi_file *mount =
        (cdi_file*)calloc(1, sizeof(cdi_file));
    if (!mount)
        err(1, "failed memory allocation");

    FILE *stream = fopen(path, "rb");
    if (!stream)
        err(1, "failure to open %s", path);

    mount->stream = stream;

    if (fseek(stream, -8, SEEK_END) < 0)
        err(1, "failure to seek to CDI footer");

    int32_t type_and_header_pos[2];
    if (fread(type_and_header_pos, sizeof(int32_t), 2, stream) != 2)
        err(1, "failure to read header offset");

    uint32_t ver = type_and_header_pos[0];

    switch (ver) {
    case CDI_V4:
    case CDI_V5:
        if (fseek(stream, type_and_header_pos[1], SEEK_SET) < 0)
            err(1, "failure to seek to header");
        break;
    case CDI_V6:
        if (fseek(stream, -type_and_header_pos[1], SEEK_END) < 0)
            err(1, "failure to seek to header");
        break;
    default:
        errx(1, "unrecognized CDI version %08x", (unsigned)ver);
    }

    uint16_t n_sessions;
    if (fread(&n_sessions, sizeof(n_sessions), 1, stream) != 1)
        err(1, "failure reading session count");

    mount->n_sessions = n_sessions;

    if (!n_sessions)
        errx(1, "CDI has zero sessions!?!?!?!?");

    mount->sessions = calloc(n_sessions, sizeof(struct cdi_session));
    if (!mount->sessions)
        err(1, "failled calloc");

    size_t total_pos = 0;
    unsigned sess_no;
    unsigned total_n_tracks = 0;
    for (sess_no = 0; sess_no < n_sessions; sess_no++) {
        read_session(stream, mount->sessions + sess_no, &total_pos, ver);
        mount->sessions[sess_no].first_track = total_n_tracks;
        total_n_tracks += mount->sessions[sess_no].n_tracks;

        unsigned n_bytes;
        if (ver == CDI_V4) {
            n_bytes = 2;
        } else {
            /*
             * TODO: THIS IS ONLY CONFIRMED TO BE THE CORRECT
             *       AMOUNT FOR CDI_V6.  I NEED TO OBTAIN A
             *       VERSION CDI_V5 CDI IMAGE FOR TESTING AND
             *       REVERSE ENGINEERING.
             */
            n_bytes = 3;
        }
        if ((sess_no < n_sessions - 1) &&
            fseek(stream, n_bytes, SEEK_CUR) < 0) {
            errx(1, "failure to seek to session");
        }
    }

    return mount;
}

static void read_session(FILE* stream,
                         struct cdi_session *sess, size_t *total_pos,
                         unsigned ver) {
    // get number of tracks, otherwise skip session structure
    int16_t n_tracks;
    if (fread(&n_tracks, sizeof(n_tracks), 1, stream) != 1)
        err(1, "failure to read session length");

    // TODO: continue from here

    if (ver == CDI_V5) {
        uint8_t len;
        if (fread(&len, sizeof(len), 1, stream) != 1)
            err(1, "failure reading session metadata");
        if (len != 0) {
            if (fseek(stream, 17, SEEK_CUR) != 0)
                err(1, "failure reading session metadata");
        } else {
            if (fseek(stream, 9, SEEK_CUR) != 0)
                err(1, "failure reading session metadata");
        }
    } else if (fseek(stream, 10, SEEK_CUR) != 0) {
        err(1, "failure reading session metadata");
    }

    if (n_tracks < 1)
        errx(1, "session has %u tracks", (int)n_tracks);

    sess->n_tracks = n_tracks;
    sess->tracks =
        (struct cdi_track*)calloc(n_tracks, sizeof(struct cdi_track));
    if (!sess->tracks)
        err(1, "failure to callco");
    unsigned track_no;
    for (track_no = 0; track_no < sess->n_tracks; track_no++)
        read_track(stream, sess->tracks + track_no, total_pos, ver);

    for (track_no = 1; track_no < sess->n_tracks; track_no++)
        if (sess->tracks[track_no - 1].sector_sz != sess->tracks[track_no].sector_sz) {
            errx(1, "not supported: discs where not all tracks in a given "
                 "session have the same sector length");
        }
}

static void read_track(FILE *stream,
                       struct cdi_track *track,
                       size_t *total_pos, unsigned ver) {
    static uint8_t const start_pattern_expect[14] =
        { 255, 255, 255, 255, 0, 0, 1, 0, 0, 0, 255, 255, 255, 255 };

    static uint8_t start_pattern[14];
    if (fread(start_pattern, sizeof(start_pattern), 1, stream) != 1)
        err(1, "failure to read track start pattern");

    if (memcmp(start_pattern, start_pattern_expect,
               sizeof(start_pattern)) != 0) {
        errx(1, "unrecognizable track start pattern!");
    }

    uint8_t path_len;
    if (fseek(stream, 4, SEEK_CUR) < 0 ||
        fread(&path_len, sizeof(path_len), 1, stream) != 1)
        err(1, "failure to read track metadata");

    unsigned seek_amt = path_len;
    if (ver == CDI_V6)
        seek_amt += 33;
    else if (ver == CDI_V4 || ver == CDI_V5)
        seek_amt += 25;
    else
        errx(1, "recognized CDI version %08x", (unsigned)ver);

    uint32_t pregap_len, track_len, start_lba, total_len;
    if (fseek(stream, seek_amt, SEEK_CUR)                 != 0 ||
        fread(&pregap_len, sizeof(pregap_len), 1, stream) != 1 ||
        fread(&track_len, sizeof(track_len), 1, stream)   != 1 ||
        fseek(stream, 22, SEEK_CUR)                        < 0 ||
        fread(&start_lba, sizeof(start_lba), 1, stream)   != 1 ||
        fread(&total_len, sizeof(total_len), 1, stream)   != 1) {
        err(1, "failure reading track metadata");
    }

    if (fseek(stream, 16, SEEK_CUR) != 0)
        err(1, "failure to seek");

    uint8_t sector_tp;
    unsigned sector_sz;
    if (fread(&sector_tp, sizeof(sector_tp), 1, stream) != 1)
        err(1, "failure reading sector type");
    switch (sector_tp) {
    case 2:
        sector_sz = 2352;
        break;
    case 1:
        sector_sz = 2336;
        break;
    default:
        err(1, "unrecognized sector type %02x", sector_tp);
    }

    uint8_t ctrl;
    if (fseek(stream, 3, SEEK_CUR) < 0 ||
        fread(&ctrl, sizeof(ctrl), 1, stream) != 1) {
        err(1, "failure to read track metadata");
    }

    switch (ver) {
    case CDI_V4:
        seek_amt = 38;
        break;
    case CDI_V5:
        /***********************************************************************
         **
         ** oh my fucking god this shit
         ** different versions of DiscJuggler would wildly alter the
         ** format of CDI images, sometimes without even changing the
         ** version ID.  There are many unknown fields, some of which
         ** are entirely arbitrary padding.  There are also several
         ** fields with variable length.
         **
         ** I don't have any info on how CDI is actually supposed to
         ** work so there's a lot of magical heuristics to figure out
         ** how far to seek, expecially in this code block.
         **
         ** The good news is that nobody actually uses The One True DiscJuggler
         ** any more so newer homebrews do actually have a fairly
         ** predictable structure, and the .cdi files are often made
         ** with open-source programs.
         **
         ** TODO: is it possible that CDI_V6 has the same
         ** behavior and i just never noticed because all the
         ** images i have on hand use the extra padding???
         **
         **********************************************************************/
        {
            seek_amt = 47;

            seek_amt -= 8;
            uint32_t wtf[2];
            if (fread(wtf, sizeof(wtf), 1, stream) != 1)
                err(1, "read failure");

            if (wtf[1] != total_len) {
                warnx("%s - inconsistent track length!  is this file "
                      "corrupt???\n", __func__);
                /*
                 * whatever try to keep going anyways even though we
                 * know we'll probably regret it...
                 */
            }

            // seak past 25 bytes of unknown data and/or padding
            if (fseek(stream, 25, SEEK_CUR) < 0)
                err(1, "seek failure");
            seek_amt -= 25;

            /*
             * i have no idea how the fuck this is supposed to
             * work im just noticing that some files have extra
             * padding and when they do there's always a four-byte
             * 0xff here.
             */
            uint32_t ohmygod;
            if (fread(&ohmygod, sizeof(ohmygod), 1, stream) != 1)
                err(1, "read failure");
            seek_amt -= 4;

            if (ohmygod == ~0)
                seek_amt += 0x7d - 47;
        }
        break;
    case CDI_V6:
    default:
        seek_amt = 0x7d;
    }

    if (fseek(stream, seek_amt, SEEK_CUR) != 0)
        err(1, "seek failure");

    track->start_lba = start_lba;
    track->start = *total_pos;
    track->pregap_len = pregap_len;
    track->track_len = track_len;
    track->sector_sz = sector_sz;
    track->total_len = total_len;
    track->ctrl = ctrl;

    *total_pos += (size_t)total_len * sector_sz;
}

static unsigned cdi_lba_base(void *ctxp) {
    cdi_file *cdi = ctxp;
    if (cdi->n_sessions >= 2 && cdi->sessions[1].n_tracks >= 1)
        return cdi->sessions[1].tracks[0].start_lba;
    else
        errx(1, "this is not a bootable MIL-CD");
}

static unsigned cdi_sector_size(void *ctxp) {
    cdi_file *cdi = ctxp;
    if (cdi->n_sessions >= 2 && cdi->sessions[1].n_tracks >= 1)
        return cdi->sessions[1].tracks[0].sector_sz;
    else
        errx(1, "this is not a bootable MIL-CD");
}

static int cdi_read_block(void *ctxp, int block, void *buf, size_t buflen) {
    cdi_file const *cdi = ctxp;

    unsigned track_no, session_no;
    for (session_no = 0; session_no < cdi->n_sessions; session_no++) {
        struct cdi_session *sess = cdi->sessions + session_no;
        for (track_no = 0; track_no < sess->n_tracks; track_no++) {
            struct cdi_track *track = sess->tracks + track_no;
            if (block >= track->start_lba && block <
                track->start_lba + track->track_len) {

                size_t byte_offset = (size_t)track->start +
                    (size_t)track->pregap_len * track->sector_sz +
                    (size_t)(block - track->start_lba) * track->sector_sz + 8;

                memset(buf, 0, buflen);

                size_t n_bytes = buflen < track->sector_sz ? buflen : track->sector_sz;

                if (fseek(cdi->stream, byte_offset, SEEK_SET) != 0) {
                    err(1, "failure to seek to track (byte offset %llx)",
                        (unsigned long long)byte_offset);
                }
                if (fread(buf, n_bytes, 1, cdi->stream) != 1) {
                    err(1, "Failure to read from cdi file");
                }
                return 0;
            }
        }
    }
    errx(1, "unable to locate LBA %u", block);
}
