/***************************************************************************
 *   Copyright (C) 2017, 2024 snickerbockers                               *
 *   snickerbockers@washemu.org                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <err.h>
#include <stdio.h>
#include <stdlib.h>

#include <glib.h>

#include "gdi_file.h"
#include "isofs.h"

static gdi_track *gdi_get_track(gdi_info *outp, int trackno);
static unsigned gdi_sector_size(void *ctxp);
static void parse_gdi(gdi_info *outp, char const *path);
static int gdi_read_block(void *ctxp, int block, void *buf, size_t buflen);
static unsigned gdi_lba_base(void *ctxp);

struct disc_intf const gdi_intf = {
    .sector_size = gdi_sector_size,
    .read_block = gdi_read_block,
    .base_lba = gdi_lba_base
};


/* enforce sane limits */
#define MAX_TRACK_FIELDS 16

gdi_info *gdi_load(char const *path) {
    gdi_info *gdi = malloc(sizeof(gdi_info));
    if (gdi)
        parse_gdi(gdi, path);
    else
        err(1, "failure to malloc");
    return gdi;
}

void parse_gdi(gdi_info *outp, char const *path) {
    long file_sz;

    FILE *stream = fopen(path, "r");

    if (!stream)
        err(1, "failure to open \"%s\"", path);

    if (fseek(stream, 0, SEEK_END) != 0)
        err(1, "While loading gdi file \"%s\"", path);

    if ((file_sz = ftell(stream)) < 0)
        err(1, "While loading gdi file \"%s\"", path);

    if (fseek(stream, 0, SEEK_SET) != 0)
        err(1, "While loading gdi file \"%s\"", path);

    /* room to append a terminator at the end just in case */
    long buf_sz = file_sz + 1;
    char *buf = (char*)g_malloc(buf_sz);

    if (fread(buf, sizeof(char), file_sz, stream) != file_sz)
        errx(1, "Failed to read from gdi file \"%s\"", path);
    buf[buf_sz - 1] = '\0';

    char **lines = g_strsplit_set(buf, "\n", -1);

    if (!lines[0])
        errx(1, "Malformed gdi file \"%s\"\n", path);

    outp->n_tracks = atoi(lines[0]);

    outp->tracks = (gdi_track*)g_malloc_n(outp->n_tracks, sizeof(gdi_track));
    char *gdi_dir = g_path_get_dirname(path);

    unsigned track_no;
    for (track_no = 1; track_no <= outp->n_tracks; track_no++) {
        if (!lines[track_no])
            err(1, "gdi file does not list all tracks");
        gdi_track *cur_track = gdi_get_track(outp, track_no);
        char **track_info = g_strsplit_set(lines[track_no], " \t",
                                           MAX_TRACK_FIELDS);

        /* maybe it's legal for them to be out of order, IDK... */
        if (!(track_info[0] && (atoi(track_info[0]) == track_no))) {
            errx(1, "Malformed gdi file \"%s\": missing track %d\n",
                 path, track_no);
        }

        if (!track_info[1] || !track_info[2] ||
            !track_info[3] || !track_info[4]) {
            errx(1, "Malformed gdi file \"%s\": track %u corrupted\n",
                 path, track_no);
        }

        cur_track->lba_start = atoi(track_info[1]);
        cur_track->ctrl = atoi(track_info[2]);
        cur_track->sector_size = atoi(track_info[3]);
        cur_track->path = g_build_filename(gdi_dir, track_info[4], NULL);

        if (track_no > 1 && cur_track->sector_size !=
            gdi_get_track(outp, track_no - 1)->sector_size) {
            errx(1, "not supported: discs where not all tracks have the same "
                 "sector length");
        }

        if (!(cur_track->stream = fopen(cur_track->path, "rb")))
            err(1, "failure to open %s", cur_track->path);

        g_strfreev(track_info);
    }

    g_free(gdi_dir);
    g_strfreev(lines);
}

// get track, trackno should be 1-indexed
static gdi_track *gdi_get_track(gdi_info *outp, int trackno) {
    return outp->tracks + (trackno - 1);
}

static unsigned gdi_sector_size(void *ctxp) {
    return ((gdi_info*)ctxp)->tracks[GDI_FIRST_DATA_TRACK - 1].sector_size;
}

static int gdi_read_block(void *ctxp, int block, void *buf, size_t buflen) {
    gdi_info *gdi = ctxp;
    gdi_track *trackp;

    // find track
    unsigned track_no;
    for (track_no = GDI_FIRST_DATA_TRACK;
         track_no <= gdi->n_tracks; track_no++) {
        if (block >= gdi_get_track(gdi, track_no)->lba_start &&
            block < gdi_get_track(gdi, track_no + 1)->lba_start)
            break;
    }

    if (track_no > gdi->n_tracks)
        track_no = gdi->n_tracks;

    trackp = gdi_get_track(gdi, track_no);
    off_t off = (block - trackp->lba_start) * trackp->sector_size;
    if (fseek(trackp->stream, off + 16, SEEK_SET) == -1)
        err(1, "failed fseek while fetching block %d", block);

    memset(buf, 0, buflen);
    if (buflen > trackp->sector_size)
        buflen = trackp->sector_size;
    if (fread(buf, buflen, 1, trackp->stream) != 1) {
        err(1, "failure to read %u bytes from block %d",
            (unsigned)buflen, block);
    }

    return 0;
}

static unsigned gdi_lba_base(void *ctxp) {
    gdi_info *gdi = ctxp;
    if (gdi->n_tracks >= GDI_FIRST_DATA_TRACK)
        return gdi_get_track(gdi, GDI_FIRST_DATA_TRACK)->lba_start;
    else
        errx(1, "no data track %u", GDI_FIRST_DATA_TRACK);
}
